/*
 	File:		ColorMatchingUtilities.c
 
 	Contains:	QuickDraw GX to PostScript conversion code.
 						File contains code necessary for supporting
						the Color Matching Utilities (Monet) through 
						the Imaging Engine.
 
 	Version:	Technology:	Quickdraw GX 1.1.x
 	 
 	Copyright:	 1990-1997 by Apple Computer, Inc., all rights reserved.
*/


#include "Traps.h"
#include "OSUtils.h"
#include "graphics types.h"
#include "graphics Routines.h"
#include "Private.h"



//<FF>
/************************************

	Routine CMUConvertColor:
	
	Routine is called instead of ConvertColor when
	a profile contains a monet tag.
	
	This is not really efficient for doing lots of colors
	but Monet isn't really geared for the needs of the imaging
	Engine.  It would be a lot of work to do this more efficiently
	in the Imaging Engine.  It would be better if Monet didn't
	require a handle for the profile, so we could just pass 
	it the result of GetTagStructure.
	
*************************************/
OSErr _CMUConvertColor(TIEGlobalsHdl hGlobals, color *target, colorSpace space)
	{
		OSErr							status;
		tag								profTag;
		CMProfileHandle		hCMProf;
		CWorld						session;
		CMColor						tempColor;
		CMMatchFlags			flags;

		GetColorProfileTags(target->profile, kMonetProfileTag, 1, 1, &profTag);
		
		status = PSSetWorkSpaceSize(hGlobals, GetTag(profTag, nil, nil));
		nrequire(status, failed_WorkSpace);
	
		hCMProf = (CMProfileHandle)(*hGlobals)->hWorkSpace;
		
		GetTag(profTag, nil, *hCMProf);

		status = CWNewColorWorld(&session, hCMProf, nil, flags);			// should get profile, not defined yet.
		nrequire(status, failed_NewSession);
		
		/*  get the color into RGB space,  use own profile to prevent matching. */
		ConvertColor(target, rgbSpace, nil, target->profile);
		
		tempColor.rgb.red = target->element.rgb.red;
		tempColor.rgb.green = target->element.rgb.green;
		tempColor.rgb.blue = target->element.rgb.blue;
		
		status = CWMatchColors(session, &tempColor, 1);
		nrequire(status = CWMatchColors(session, &tempColor, 0), failed_Match);

		switch(space) {
		
			case rgbSpace:
				target->element.rgb.red = tempColor.rgb.red;
				target->element.rgb.green = tempColor.rgb.green;
				target->element.rgb.blue = tempColor.rgb.blue;
				
				break;
				
			case graySpace:
				target->element.gray = ( 30 * tempColor.rgb.red +
																 59 * tempColor.rgb.green +
																 11 * tempColor.rgb.blue ) / 100;
				
				break;
				
			case cmykSpace:
				{
					colorValue			min;

					target->element.cmyk.cyan = tempColor.rgb.red ^ 0xFFFF;
					min = target->element.cmyk.cyan;

					target->element.cmyk.magenta = tempColor.rgb.green ^ 0xFFFF;
					if (min > target->element.cmyk.magenta)
						min = target->element.cmyk.magenta;

					target->element.cmyk.yellow = tempColor.rgb.blue ^ 0xFFFF;
					if (min > target->element.cmyk.yellow)
						min = target->element.cmyk.yellow;
					
					target->element.cmyk.black = min;
					target->element.cmyk.cyan -= min;
					target->element.cmyk.magenta -= min;
					target->element.cmyk.yellow -= min;
					
				}
				break;
				
			}//end switch
			
			CWDisposeColorWorld(session);

failed_Match:
failed_NewSession:	
		{
			register OSErr saveStatus = PSReleaseWorkSpace(hGlobals);
			ncheck(saveStatus);
			
			if (status == noErr)
				status = saveStatus;
				
		}//end block
failed_WorkSpace:	
		return(status);
	
	}//CMUConvertColor



//<FF>
/************************************

	Routine CMUMatchBitmap:
	
	Routine takes an rgb32 bitmap and matches
	it using Monet and converts it into the destination
	space (only CMYK or RGB are allowed)


**************************************/
OSErr _CMUMatchBitmap(TIEGlobalsHdl hGlobals, bitmap *theBitmap, colorSpace destSpace)
	{
		OSErr							status;
		tag								profTag;
		CMProfileHandle		hCMProf;
		CMMatchFlags			flags;
		PixMap						thePixMap;
		unsigned long			*rgb32Pixel, cmyk32Value;
		long							i, j, rowLongs;
		Ptr								pRaster;
		unsigned char			min, *cmyk;
	
		check( theBitmap->space == rgb32Space);
	
		GetColorProfileTags(theBitmap->profile, kMonetProfileTag, 1, 1, &profTag);
		
		status = PSSetWorkSpaceSize(hGlobals, GetTag(profTag, nil, nil));
		nrequire(status, failed_WorkSpace);
	
		hCMProf = (CMProfileHandle)((*hGlobals)->hWorkSpace);
		
		GetTag(profTag, nil, *hCMProf);	

		thePixMap.baseAddr = theBitmap->image;
		thePixMap.rowBytes = theBitmap->rowBytes | 0x8000;
		thePixMap.bounds.top = 0;
		thePixMap.bounds.right = theBitmap->width - 1;
		thePixMap.bounds.left = 0;
		thePixMap.bounds.bottom = theBitmap->height - 1;
		thePixMap.packType = 0;
		thePixMap.pixelType = chunky;
		thePixMap.pixelSize = 32;
		thePixMap.cmpCount = 3;
		thePixMap.cmpSize = 8;
		thePixMap.pmTable = nil;
		
		status = MatchPixMap(hCMProf, nil, flags, &thePixMap , nil);						// Should get profile, not defined yet.
		nrequire(status, failed_Match);
		
		/** Now convert the bitmap into cmyk if necessary **/
		
		if (destSpace == cmykSpace) {
			
			rowLongs = theBitmap->rowBytes / 4;					// must be long alligned since it is 32 bit.
			pRaster = theBitmap->image;
			for (i = 0; i < theBitmap->height; i++) {
		
				rgb32Pixel = (unsigned long*)pRaster;
				for ( j = 0; j < rowLongs; j++) {
				
					cmyk32Value = (*rgb32Pixel ^ 0xFFFFFFFF) << 8;	// invert components, shift up to cmy, leaves K 0;
					
					cmyk = (unsigned char*)(&cmyk32Value);					// point cmyk byte array at the long.
					
					/** Do full under color removal **/
					
					min = cmyk[0];
					if (min > cmyk[1])
						min = cmyk[1];
					if (min > cmyk[2])
						min = cmyk[2];
						
					cmyk[3] = min;
					cmyk[2] -= min;
					cmyk[1] -= min;
					cmyk[0] -= min;
					
					*rgb32Pixel++ = cmyk32Value;
				
				}//end for
				
				pRaster += theBitmap->rowBytes;
		
			}//end for
		
		}//end if
		
failed_Match:
failed_WorkSpace:

		{
			register OSErr saveStatus = PSReleaseWorkSpace(hGlobals);
			ncheck(saveStatus);
			
			if (status == noErr)
				status = saveStatus;
				
		}//end block

		return(status);
		
	}//CMUMatchBitmap



//<FF>
/********************************
	Function TestforCMU:
	
	Function returns whether or not the 
	color matching utilities are present.
	
*********************************/
Boolean _TestForCMU()
	{

#if BUILDING_FOR_NUKERNEL

		// Always true on Maxwell.
		
		return (true);

#else

		long				uTrap;
		long				theTrap;
		
		theTrap = GetTrapAddress(CMTrap);
		uTrap = GetTrapAddress(_Unimplemented);
				
		if (uTrap == theTrap)
			return(false);
		else
			return(true);

#endif
	
	}//TestForCMU

